/*
 * Copyright 2008, Petr Machata
 *
 * This file is part of MicroSweeper.
 *
 * MicroSweeper is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MicroSweeper is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MicroSweeper.  If not, see <http://www.gnu.org/licenses/>.
 */

public abstract class MineField {

    protected final Field[] fields;
    protected final ParticipantSelector parSel;

    private boolean busted = false;
    private boolean won = false;

    protected MineField(int numberOfFields, ParticipantSelector parSel)
	throws IllegalArgumentException
    {
	this.fields = new Field[numberOfFields];
	for (byte i = 0; i < numberOfFields; ++i)
	    fields[i] = new NormalField(i);
	this.parSel = parSel;
    }

    public MineField(MineField copy) {
	int s = copy.fields.length;
	fields = new Field[s];
	for (int i = 0; i < s; ++i)
	    fields[i] = copy.fields[i].clone();
	parSel = copy.parSel;
    }

    abstract public MineField clone();

    // - - - - - - - - - - - - - - - - -
    // Coordinate handling
    // - - - - - - - - - - - - - - - - -

    // Number of neighbors of each field--i.e. fields that one can
    // move to from current field.
    abstract public byte getCountNeighborIndices();

    // mandatory directions
    abstract public byte getDirUp();
    abstract public byte getDirDown();
    abstract public byte getDirLeft();
    abstract public byte getDirRight();

    abstract public byte getReciprocal(byte dir1);

    // return an index, or -1 if there's no neighbor in that direction
    abstract public byte getNeighbor(byte fieldIdx, byte dir1);

    // where should the cursor be placed when the game starts
    abstract public byte getInitialPos();

    public byte getNumberOfFields() {
	return (byte)fields.length;
    }

    // Useful for MineField subclasses that have a notion of width and
    // height.
    protected int boundWrap(int c, int s) {
	if (c < 0)
	    return s + c % s;
	else if (c >= s)
	    return c % s;
	else
	    return c;
    }

    protected int bound(boolean wrap, int c, int s) {
	if (wrap)
	    return boundWrap(c, s);
	else {
	    if (c < 0)
		return 0;
	    else if (c >= s)
		return s - 1;
	    else
		return c;
	}
    }


    // - - - - - - - - - - - - - - - - -
    // Field handling
    // - - - - - - - - - - - - - - - - -

    public Field getField(byte fieldIdx) {
	return fields[fieldIdx];
    }

    public void putField(byte fieldIdx, Field field) {
	System.out.println("putField...");
	fields[fieldIdx] = field;
	System.out.println("done.");
    }

    public Field getNeighbor(Field f, byte dir1) {
	byte fieldIdx2 = getNeighbor(f.fieldIdx, dir1);
	if (fieldIdx2 < 0)
	    return null;
	else
	    return getField(fieldIdx2);
    }

    public void allFields(FieldEnumerator e) {
	for (int i = 0; i < fields.length; ++i)
	    e.field(fields[i]);
    }

    public void allNeighbors(byte fieldIdx, FieldEnumerator e) {
	byte count = getCountNeighborIndices();
	for (byte nIdx = 0; nIdx < count; ++nIdx) {
	    byte fieldIdx2 = getNeighbor(fieldIdx, nIdx);
	    if (fieldIdx2 < 0)
		continue;
	    if (!e.field(getField(fieldIdx2)))
		break;
	}
    }

    public boolean uncover(byte idx1)
	throws IllegalArgumentException
    {
	switch (fields[idx1].uncover()) {
	case Field.UNCOVER_BUSTED:
	    busted = true;
	    return true;
	case Field.UNCOVER_NEW:
	    checkWin();
	    return true;
	}
	return false;
    }

    // - - - - - - - - - - - - - - - - -
    // Participants
    // - - - - - - - - - - - - - - - - -

    // Participant enumerator wrapper.
    public void allParticipants(byte fieldIdx, FieldEnumerator e) {

	byte count = parSel.getCountParticipantIndices(this);
	for (byte parIdx = 0; parIdx < count; ++parIdx) {
	    byte fieldIdx2 = parSel.getParticipant(this, fieldIdx, parIdx);
	    if (fieldIdx2 < 0)
		continue;
	    if (!e.field(this.getField(fieldIdx2)))
		break;
	}
    }

    // Answers whether fieldIdx1 is participant of fieldIdx2
    public boolean isParticipantOf(final byte fieldIdx1, byte fieldIdx2) {

	class E implements FieldEnumerator {
	    public boolean found = false;
	    public boolean field(Field f) {
		if (f.fieldIdx == fieldIdx1)
		    found = true;
		return !found;
	    }
	}
	E e = new E ();
	allParticipants (fieldIdx2, e);
	return e.found;
    }

    public byte countParticipantPred(byte fieldIdx, final FieldEnumerator pred) {

	class E implements FieldEnumerator {
	    byte count = 0;
	    public boolean field(Field f) {
		if (pred.field(f)) {
		    ++count;
		}
		return true;
	    }
	}
	E e = new E();
	allParticipants(fieldIdx, e);
	return e.count;
    }

    public byte countParticipantMines(byte idx1) {
	return countParticipantPred(idx1, new FieldEnumerator() {
		public boolean field(Field f) {
		    return f.mine;
		}
	    });
    }

    public void updateParticipantCounts() {
	for (byte fieldIdx = 0; fieldIdx < fields.length; ++fieldIdx)
	    fields[fieldIdx].setParCount
		(countParticipantMines(fieldIdx));
    }

    // - - - - - - - - - - - - - - - - -
    // Game state
    // - - - - - - - - - - - - - - - - -

    public boolean isBusted() {
	return busted;
    }

    public boolean hasWon() {
	return won;
    }

    public boolean gameOver() {
	return hasWon() || isBusted();
    }

    private void checkWin() {
	if (busted)
	    return;
	for (byte idx1 = 0; idx1 < fields.length; ++idx1) {
	    boolean mine = fields[idx1].mine;
	    boolean uncovered = fields[idx1].uncovered;
	    if (mine && uncovered)
		return;
	    if (!mine && !uncovered)
		return;
	}
	won = true;
    }

    // - - - - - - - - - - - - - - - - -
    // Encoding
    // - - - - - - - - - - - - - - - - -

    public String getEncoding() {
	// Will fail loudly if fields.length % 5 != 0
	StringBuffer buf = new StringBuffer();
	for (byte idx1 = 0; idx1 < fields.length; ) {
	    boolean a1 = fields[idx1++].mine;
	    boolean a2 = idx1 < fields.length ? fields[idx1++].mine : false;
	    boolean a3 = idx1 < fields.length ? fields[idx1++].mine : false;
	    boolean a4 = idx1 < fields.length ? fields[idx1++].mine : false;
	    boolean a5 = idx1 < fields.length ? fields[idx1++].mine : false;
	    int code = (a1 ? 16 : 0) + (a2 ? 8 : 0)
		+ (a3 ? 4 : 0) + (a4 ? 2 : 0) + (a5 ? 1 : 0);
	    char c = "abcdefghijklmnop".charAt(code % 16);
	    if (code > 15)
		c = Character.toUpperCase(c);
	    buf.append(c);
	}
	buf.append(';');
	return buf.toString();
    }

    public int hashCode() {
	int hash = 0;
	for (byte idx1 = 0; idx1 < fields.length; ++idx1) {
	    int h0 = ((hash >>> 24) & 0xff);
	    hash = hash << 1;
	    hash = hash ^ h0;
	    if (fields[idx1].mine)
		hash |= 1;
	    else
		hash &= ~1;
	}
	return hash;
    }

    public void paint(final MineSweeperDisplay d) {
	d.beforeDrawing();
	allFields(new FieldEnumerator() {
		public boolean field(Field f) {
		    paintField(f, d);
		    return true;
		}
	    });
	d.afterDrawing();
    }

    public abstract void paintField(Field f, MineSweeperDisplay d);

    public void paint(boolean full) {
	paint(new StringGridDisplay(System.out, full));
    }

    public void paint() {
	paint(true);
    }
}
