/*
 * Copyright 2008,2009,2010 Petr Machata
 *
 * This file is part of MicroSweeper.
 *
 * MicroSweeper is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MicroSweeper is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MicroSweeper.  If not, see <http://www.gnu.org/licenses/>.
 */

import javax.microedition.lcdui.*;
import javax.microedition.midlet.*;
import java.lang.Math;

public class MineSweeperCanvas
    extends Canvas
    implements CommandListener
{
    final Font statusBarFont = Font.getFont(Font.FACE_MONOSPACE,
					    Font.STYLE_PLAIN,
					    Font.SIZE_MEDIUM);

    private final MicroSweeper parent;
    private boolean building = false;

    private final MineField mineField;
    private final MineFieldBuilder builder;

    private final Command backCommand;
    private final Command flagCommand;
    private final Command restartCommand;

    // Start at the corner, where there is the best chance to get a
    // cascade on the first click.
    private byte cursor;
    private int flagsLeft;
    private boolean showParticipantHints = true;

    private class FontBestMatch {
	private int diff = 0;
	private Font bestMatch = null;
	final private int wantHeight;

	private void checkBestMatch(int face, int style, int size) {
	    Font font = Font.getFont(face, style, size);
	    int baseline = font.getBaselinePosition();
	    int d = baseline - wantHeight;

	    if (d > 0)
		d *= 5; // Penalize too large a font, but still select
			// it if tiny font is the only alternative.

	    if (bestMatch == null || Math.abs(d) < Math.abs(diff))
		bestMatch = font;
	}

	public FontBestMatch(int wantHeight) {
	    this.wantHeight = wantHeight;
	}

	public FontBestMatch checkFont(int face, int style) {
	    checkBestMatch(face, style, Font.SIZE_SMALL);
	    checkBestMatch(face, style, Font.SIZE_MEDIUM);
	    checkBestMatch(face, style, Font.SIZE_LARGE);
	    return this;
	}

	public Font getBestMatch() {
	    return bestMatch;
	}
    }

    public MineSweeperCanvas(MicroSweeper parent, MineFieldBuilder builder) {
	this.parent = parent;
	this.builder = builder;
	this.mineField = builder.mineField;
	this.flagsLeft = builder.getNumMines();
	this.cursor = mineField.getInitialPos();

	addCommand(backCommand = new Command("Back", Command.BACK, 0));
	addCommand(flagCommand = new Command("Flag", Command.ITEM, 0));
	restartCommand = new Command("Restart", Command.ITEM, 1);
	setCommandListener(this);
    }

    private int getR() {
	int w = getWidth();
	int sbs = statusBarFont.getBaselinePosition();
	int h = getHeight() - sbs;
	return w < h ? w : h;
    }

    private class CanvasDisplay
	implements MineSweeperDisplay
    {
	protected final int r;    // The side of the square that contains the mine field.
	protected final Graphics g;
	protected final boolean isColor;

	public CanvasDisplay(Graphics g) {
	    this.g = g;
	    this.r = getR();

	    // See if we have colors.
	    Display disp = Display.getDisplay(parent);
	    this.isColor = disp.isColor();
	}

	protected boolean shouldDrawBuilderProgress(byte idx1, int progress) {
	    return progress > idx1;
	}

	private void paintFlag(int x0, int y0, int x1, int y1) {

	    int w0 = x1 - x0 + 1;
	    int h0 = y1 - y0 + 1;

	    g.setColor(0);
	    int xm = (x0 + x1) / 2;
	    g.drawLine(xm - 2, y0 + 5, xm - 2, y1 - 5);
	    g.drawLine(xm - 1, y0 + 5, xm - 1, y1 - 5);
	    g.drawLine(xm, y0 + 5, xm, y1 - 5);

	    if (Display.getDisplay(parent).isColor())
		g.setColor(0xff0000);
	    else
		g.setColor(0xffffff);
	    g.fillRect(xm, y0 + 5, w0 / 3, h0 / 4);
	}

	private void paintCross(int x0, int y0, int x1, int y1) {

	    if (Display.getDisplay(parent).isColor())
		g.setColor(0xff0000);
	    else
		g.setColor(0);

	    g.drawLine(x0+2, y0+2, x1-2, y1-2);
	    g.drawLine(x0+3, y0+2, x1-2, y1-3);
	    g.drawLine(x0+2, y0+3, x1-3, y1-2);

	    g.drawLine(x0+2, y1-2, x1-2, y0+2);
	    g.drawLine(x0+3, y1-2, x1-2, y0+3);
	    g.drawLine(x0+2, y1-3, x1-3, y0+2);
	}

	public void drawField(Field f, int fx, int fy, int fr) {

	    int x0 = this.r * fx / Constants.paintAccuracy;
	    int y0 = this.r * fy / Constants.paintAccuracy;
	    int w0 = (this.r * (fx + fr) / Constants.paintAccuracy) - x0;
	    int h0 = (this.r * (fy + fr) / Constants.paintAccuracy) - y0;

	    doDrawField(f, x0, y0, w0, h0);
	}

	protected void doDrawField(Field f, int x0, int y0, int w0, int h0) {

	    int x1 = x0 + w0 - 1;
	    int y1 = y0 + h0 - 1;

	    // @todo cache this, it's not going to change anyway...
	    Font mineCountFont = new FontBestMatch(h0)
		.checkFont(Font.FACE_MONOSPACE, Font.STYLE_BOLD)
		.getBestMatch();
	    g.setFont(mineCountFont);

	    int d = (h0 - mineCountFont.getBaselinePosition()) / 2;
	    int center = (x0 + x1) / 2;


	    if (!f.uncovered || building) {

		g.setColor(0x00808080);
		g.fillRect(x0, y0, w0, h0);

		// Shade...
		g.setColor(0x00505050);
		g.drawLine(x0+1, y1, x1, y1);
		g.drawLine(x1, y1, x1, y0+1);

		g.setColor(0x00707070);
		g.drawLine(x0+2, y1-1, x1-1, y1-1);
		g.drawLine(x1-1, y1-1, x1-1, y0+2);

		// Light...
		g.setColor(0x00b0b0b0);
		g.drawLine(x0, y1-1, x0, y0);
		g.drawLine(x0, y0, x1-1, y0);

		g.setColor(0x00909090);
		g.drawLine(x0+1, y1-2, x0+1, y0+1);
		g.drawLine(x0+1, y0+1, x1-2, y0+1);

		if (building
		    && shouldDrawBuilderProgress
			(f.fieldIdx, builder.getProgressMines()))
		    paintFlag(x0, y0, x1, y1);
		else if (f.flagged) {
		    if (mineField.gameOver() && !f.mine)
			paintCross(x0, y0, x1, y1);
		    else
			paintFlag(x0, y0, x1, y1);
		}

	    } else {

		g.setColor(0x00909090);
		g.fillRect(x0, y0, w0, h0);

		g.setColor(0x00a0a0a0);
		g.drawRect(x0, y0, w0, h0);

		if (!f.mine && !f.consistent(mineField)) {
		    if (isColor)
			g.setColor(0xff0000);
		    else
			g.setColor(0);
		    g.drawRect(x0 + 1, y0 + 1, w0 - 2, h0 - 2);
		}

		String s;
		if (!f.mine && (s = f.render()) != "") {
		    if (isColor)
			g.setColor(f.getColor());
		    else
			g.setColor(0);

		    g.drawString(s, center, y0 + d,
				 Graphics.HCENTER | Graphics.TOP);
		}
	    }

	    if (!f.flagged && f.mine
		&& (mineField.isBusted()
		    || mineField.hasWon()
		    || f.uncovered)) {

		if (mineField.hasWon())
		    paintFlag(x0, y0, x1, y1);

		else {
		    g.setColor(0);
		    g.fillArc(x0 + 5, y0 + 5, w0 - 11, h0 - 11, 0, 360);
		}
	    }

	    // Draw cursor and participant hints.
	    if (!mineField.gameOver()) {
		if (f.fieldIdx == cursor) {
		    g.setColor(0x00d0d0d0);
		    g.drawRect(x0 + 2, y0 + 2, w0 - 5, h0 - 5);
		}
		else if (showParticipantHints
			 && !f.uncovered
			 && mineField.isParticipantOf(f.fieldIdx, cursor)) {
		    g.setColor(0x00ffffff);
		    g.drawRect(x0 + w0 / 2 - 1, y0 + h0 / 2 -1, 3, 3);
		}
	    }
	}

	public void beforeDrawing() {
	    g.setColor(0);
	    g.fillRect(0, 0, r, r);
	}

	public void afterDrawing() {

	    int w = getWidth();
	    int h = getHeight();

	    g.setColor(0);
	    g.fillRect(0, r, r, h - r);

	    g.setColor(0);
	    g.fillRect(r, 0, w - r, h);

	    g.setColor(0xffffff);
	    g.setFont(statusBarFont);
	    g.drawString(mineField.isBusted() ? ":-("
			 : (mineField.hasWon() ? ":-)"
			    : Long.toString(flagsLeft)),
			 5, h, Graphics.LEFT | Graphics.BOTTOM);
	}
    }

    class WallPainerDisplay
	extends CanvasDisplay
    {
	public WallPainerDisplay(Graphics g) {
	    super(g);
	}

	private boolean shouldDrawWall(byte idx1, byte dir1, byte w) {

	    // Early confirmation when the building is in progress.
	    if (building
		&& shouldDrawBuilderProgress(idx1, builder.getProgressWalls()))
		return true;

	    // Easy cases: if there's no wall, don't draw it, if there is
	    // and the field is uncovered, draw it.
	    Field f = mineField.getField(idx1);
	    if (true)//!f.hasWall(w))
		return false;
	    if (f.uncovered
		|| mineField.isBusted()
		|| mineField.hasWon())
		return true;

	    // However we also need to draw a wall if it's near the
	    // uncovered field.  E.g. see following:
	    //    +---+---+---+
	    //    |   | 1 | 1 |
	    //    +---+---+---+
	    //    |   | 1 | * |
	    //    +---+===+===+
	    //    |   | 1 | * |
	    //    +---+---+---+
	    // (All fields are uncovered, except for the ones denoted with
	    // "*", which contain bombs).
	    // If the wall between the two covered fields is not drawn,
	    // the player may get the wrong impression that lower field
	    // doesn't contain a bomb.

	    class E implements FieldEnumerator {
		boolean uncoveredNeighbor = false;
		public boolean field(Field f) {
		    if (f.uncovered)
			uncoveredNeighbor = true;
		    return !uncoveredNeighbor;
		}
	    }
	    E e = new E();
	    mineField.allNeighbors(idx1, e);
	    return e.uncoveredNeighbor;
	}

	private void maybeDrawWall(byte idx1, byte dir1,
				   int x0, int y0, int x1, int y1) {

	    byte w = 0; //mineField.toWall(dir1);
	    if (shouldDrawWall(idx1, dir1, w)) {

		g.setColor(0);
		g.drawLine(x0, y0, x1, y1);

		int dx = 0; //Field.wallDx(w);
		int dy = 0; //Field.wallDy(w);
		g.drawLine(x0-dx, y0-dy, x1-dx, y1-dy);
		g.drawLine(x0+dx, y0+dy, x1+dx, y1+dy);
		g.drawLine(x0+2*dx, y0+2*dy, x1+2*dx, y1+2*dy);
	    }
	}

	protected void doDrawField(Field f, int x0, int y0, int w0, int h0) {

	    int x1 = x0 + w0 - 1;
	    int y1 = y0 + h0 - 1;

	    maybeDrawWall(f.fieldIdx, mineField.getDirLeft(), x0, y0, x0, y1);
	    maybeDrawWall(f.fieldIdx, mineField.getDirRight(), x1, y0, x1, y1);
	    maybeDrawWall(f.fieldIdx, mineField.getDirUp(), x0, y0, x1, y0);
	    maybeDrawWall(f.fieldIdx, mineField.getDirDown(), x0, y1, x1, y1);
	}

	public void beforeDrawing() {}
	public void afterDrawing() {}
    }

    class RepaintDisplay
	implements MineSweeperDisplay
    {
	public void drawField(Field f, int fx, int fy, int fr) {
	    int r0 = getR();
	    int fieldS = r0 * fr / Constants.paintAccuracy;

	    int x0 = r0 * fx / Constants.paintAccuracy;
	    int y0 = r0 * fy / Constants.paintAccuracy;

	    // repaint more to accomodate for corners
	    repaint(x0 - 3, y0 - 3,
		    fieldS + 6, fieldS + 6);
	}

	public void beforeDrawing() {}
	public void afterDrawing() {}
    }
    MineSweeperDisplay repaintDisplay = new RepaintDisplay();

    protected void paint(Graphics g) {
	// The display has to be reconstructed on each repaint,
	// because in general we don't know if graphics size didn't
	// change (plus the old reference to `g' would probably be
	// invalid).
	mineField.paint(new CanvasDisplay(g));
	mineField.paint(new WallPainerDisplay(g));
    }

    private void repaintField(Field f) {
	mineField.paintField(f, repaintDisplay);
    }

    private void repaintStatusBar() {
	int r = getR();
	repaint(0, r, getWidth(), getHeight() - r);
    }

    private boolean moveCursor(byte dir1) {
	byte nc = mineField.getNeighbor(cursor, dir1);
	if (nc < 0 || nc == cursor)
	    return false;

	FieldEnumerator participantRepainter
	    = new FieldEnumerator() {
		    public boolean field(Field f) {
			repaintField(f);
			return true;
		    }
		};

	repaintField(mineField.getField(cursor));
	mineField.allParticipants(cursor, participantRepainter);
	repaintField(mineField.getField(nc));
	mineField.allParticipants(nc, participantRepainter);

	cursor = nc;
	return true;
    }

    private boolean toggleFlag(final byte idx1) {
	Field f = mineField.getField(idx1);
	boolean hadFlag = f.flagged;
	if (flagsLeft > 0 || hadFlag) {
	    f.toggleFlag();
	    repaintField(f);
	    mineField.allParticipants
		(idx1, new FieldEnumerator()	{
			public boolean field(Field f) {
			    repaintField(f); // xxx only necessary if
					     // consistency changed
			    return true;
			}
		    });
	    if (hadFlag)
		flagsLeft += 1;
	    else if (f.flagged)
		flagsLeft -= 1;
	    repaintStatusBar();
	    return true;
	}
	else
	    return false;
    }

    private boolean uncover(final byte idx1) {
	showParticipantHints = false;

	Field f = mineField.getField(idx1);
	if (!f.uncovered && !f.flagged) {
	    mineField.uncover(idx1);

	    if (f.getParCount() == 0 && !f.mine)
		mineField.allParticipants
		    (idx1, new FieldEnumerator() {
			    public boolean field(Field f) {
				if (!f.uncovered)
				    uncover(f.fieldIdx);
				return true;
			    }
			});

	    mineField.allParticipants(idx1, new FieldEnumerator() {
		    public boolean field(Field f) {
			repaintField(f);
			return true;
		    }
		});
	    repaintField(f);
	    return true;

	} else
	    return false;
    }

    private void doneBuilding(byte idx1) {
	building = false;
	repaint();
	System.out.println("done building");
	fire(idx1);
    }

    private void fire(byte idx1) {
	uncover(idx1);
	if (mineField.isBusted() || mineField.hasWon()) {
	    repaint();
	    removeCommand(flagCommand);
	    addCommand(restartCommand);
	}
    }

    protected void keyPressed(int keyCode) {
	if (mineField.hasWon() || mineField.isBusted() || building)
	    return;

	switch (getGameAction(keyCode)) {
	case UP:
	    moveCursor(mineField.getDirUp());
	    break;

	case DOWN:
	    moveCursor(mineField.getDirDown());
	    break;

	case LEFT:
	    moveCursor(mineField.getDirLeft());
	    break;

	case RIGHT:
	    moveCursor(mineField.getDirRight());
	    break;

	case FIRE:
	    if (!builder.isInitialized()) {
		building = true;

		class RepaintThread extends Thread {
		    private boolean done = false;
		    public void doneBuilding() {
			done = true;
		    }
		    public void run() {
			int last = 0;
			while (!done) {
			    try {
				sleep(500);
			    } catch (InterruptedException e) {
				// never mind...
			    }
			    int prog = Math.max(builder.getProgressWalls(),
						builder.getProgressMines());
			    int max = Math.max(prog, last);
			    int min = Math.min(prog, last);
			    for (int i = min; i <= max; ++i)
				repaintField(mineField.getField((byte)i));
			    last = prog;
			}
		    }
		}

		class BuilderThread extends Thread {
		    RepaintThread rpt;
		    public BuilderThread(RepaintThread t) {
			rpt = t;
		    }
		    public void run() {
			builder.initialize(cursor);
			rpt.doneBuilding();
			MineSweeperCanvas.this.doneBuilding(cursor);
		    }
		}

		RepaintThread rpt = new RepaintThread();
		BuilderThread bt = new BuilderThread(rpt);
		rpt.start();
		bt.start();
	    } else
		fire(cursor);
	}
    }

    public void commandAction(Command c, Displayable s) {
	if (c == backCommand) {
	    if (mineField.gameOver())
		parent.exitGame();
	    else
		parent.showMenu(this);

	} else if (c == restartCommand)
	    parent.restartGame();

	else if (!mineField.gameOver()) {
	    if (c == flagCommand
		&& !mineField.getField(cursor).uncovered
		&& builder.isInitialized()) {

		toggleFlag(cursor);
	    }
	}
    }

    public void stop() {
	builder.stop();
    }
}
